import copy, time
import dxapi

# Note: install dxapi from pip nexus repository 
#       pip install --extra-index-url https://nexus.deltixhub.com/repository/epm-rtc-public-python/simple
# or local directory from QuantServer installation:
#       pip install /path/to/QuantServer/python
#    or pip install .. (from current directory)

# In the sample we will update fields in securities stream.
# The sample works with dxapi v5.4.30 and TB server 5.5.67.

def readInstruments(stream, entities):
    messages = []
    with stream.trySelect(dxapi.JAVA_LONG_MIN_VALUE, dxapi.SelectionOptions(), None, entities) as cursor:
        while cursor.next():
            message = copy.deepcopy(cursor.getMessage())
            message.typeId = None # drop typeId because we want to insert messages later, this string is not requered with dxapi v5.4.31
            messages.append(message)
    return messages

def insertInstruments(stream, messages):
    options = dxapi.LoadingOptions()
    options.writeMode = dxapi.WriteMode('INSERT')
    with stream.tryLoader(options) as loader:
        for message in messages:
            loader.send(message)

def copyStream(source, target):
    print('Copying stream ' + source.key() + ' into ' + target.key())
    count = 0
    with source.trySelect(dxapi.JAVA_LONG_MIN_VALUE, dxapi.SelectionOptions(), None, None) as cursor, target.tryLoader(dxapi.LoadingOptions()) as loader:
        while cursor.next():
            message = cursor.getMessage()
            message.typeId = None
            loader.send(message)
            count = count + 1
    print('Copied ' + str(count) + ' messages')

def backupSecuritiesStream(db):
    source = db.getStream('securities')
    name = 'securities#backup.' + str(round(time.time() * 1000))
    options = dxapi.StreamOptions()
    options.name(name)
    # For dxapi v < 5.4.28 and TB server v < 5.5.67
    # options.polymorphic = True
    # options.metadataXml(source.metadataXml())
    options.schema(source.schema())
    print('Creating backup stream: ' + name)
    target = db.createStream(name, options)
    print('OK')
    copyStream(source, target)

timebase = 'dxtick://localhost:8011'
with dxapi.TickDb.openFromUrl(timebase, False) as db:
    print('Connected to ' + timebase)

    stream = db.getStream('securities')
    if stream == None:
        raise Exception('securities stream not found, nothing to do')

    # get write lock if need
    stream.tryLock(dxapi.LockType('WRITE'), 5000)
    try:
        # List of entities to subscribe (if None, all stream entities will be used)
        entities = [
            dxapi.InstrumentIdentity(dxapi.InstrumentType.FX, 'BTC/USDT'),
            dxapi.InstrumentIdentity(dxapi.InstrumentType.FX, 'BTC/ETH')
        ]
        messages = readInstruments(stream, entities)
        print('Read ' + str(len(messages)) + ' messages')

        if len(messages) > 0:
            # Note: backup securities stream before changing, comment if you don't need
            backupSecuritiesStream(db)

            # update instruments
            for message in messages:
                # change elements
                message.makerCommission = 0.022

            # remove instruments from securties
            stream.clear(entities)
            print('Cleared ' + str(len(messages)) + ' entities')

            # add updated instruments to securities
            insertInstruments(stream, messages)
            print('Inserted ' + str(len(messages)) + ' messages')
    finally:
        stream.unlock()